#!/usr/bin/env python
 
###############################################################################
#
# PipelineUtils.py
#
# Purpose:
#
#    General puprose utilities for OPUS Python scripts.
#
# Limitations:
#
#    - checkServersUp() relies on the assumption that we are running this
#         on the same machine where the servers are running.
#
# History:
# -------- -----  --------  --------------------------------------
# 10/06/04 xxxxx  Sontag    first version
# 09/07/05 52948  Sherbert  Add 3 modules for path file reading
# 05/12/06 55843  Sherbert  Skip comment lines in pipeline file
# 05/17/06 55256  Sontag    dont want newline char from getSingleLineOutputOf()
# 02/21/07 55256  Sontag    Add some smart wrappers to pstat_test calls
# 06/20/07 55256  Sontag    Remove use of deprecated string module
# 10/20/11 67891  MSwam     Copy from AUTO to SHARE, for CDBS pipeline use
###############################################################################

import copy, os, pwd, re, sys, time
import commands, shutil

#
# precursor to "import debug"
#
# returns nothing
#
DEBUG = 0

def debug(msg):
   if (1 != DEBUG):
      return

   date = time.asctime(time.localtime(time.time()))
   datedMsg = date + " : " + msg

   print datedMsg


#
# Return the lowest byte of an integer.  This is useful
# for shifting the return value received from the os.system()
# command.
#
def byt(anInt):
   return int(anInt & 0xff)


#
# My own grep emulator, until one is built in.
# Input is a file.  Returns a list of strings.
#
def mygrep_file(pattern, filename):
   file = open(filename, 'r')
   fileLines = file.readlines()  # reads all contents into a list of strings
   file.close()
   resultList = [] # empty for now
   for line in fileLines:
      if (None != re.search(pattern, line)):
         resultList.append(line)
   return resultList


#
# My own grep emulator, until one is built in.
# Input is a multi-line string.  Returns a list of strings.
#
def mygrep_buf(pattern, buffer):
   fileLines = buffer.split("\n") # make like a list of file lines
   resultList = [] # empty for now
   for line in fileLines:
      if (None != re.search(pattern, line)):
         resultList.append(line)
   return resultList


#
# Utility routine to grab a password out of an OPUS password file.
# Assumes it is located in $HOME/.opus_keys
#
# throws IOError if file is not found
#
def getPasswordFromKey(key):
   # init
   home = os.environ["HOME"]
   datFile = home + "/.opus_keys"
   retval = ""

   # read the file
   file = open(datFile, 'r')
   fileLines = file.readlines()  # reads all contents into a list of strings
   file.close()

   # look for the key
   for line in fileLines:
      items = line.strip().split()
      if (len(items) > 1):
         if (items[0] == key):
            retval = items[1]
            return retval

   # if we get here, then we haven't found a matching key
   assert len(retval) > 0, "Key not found: "+key +", in file: "+datFile

#
# A simple wrapper to popen to avoid duplication of these steps each time 
# a system command is run.  Puts the output of the command (stdout&stderr)
# in the return string.
#
def getOutputOf(cmd):
   # run the command with popen
   cmdPipes = os.popen4(cmd) # use popen4 to grab both stdout and stderr
   r0 = cmdPipes[0].close()
   cmdOut = cmdPipes[1].read()
   r1 = cmdPipes[1].close()
#  args.append(r1) # unfortunately, exit status isn't getting set as r1 or r2 !
   if (len(cmdOut) > 0 and cmdOut[-1] == '\n'):
      cmdOut = cmdOut[0:-1]
   return cmdOut

# Same as above, but return as a list of lines
def getListOutputOf(cmd):
   buf = getOutputOf(cmd)
   theList = buf.split("\n")
   while (theList.count('') > 0): theList.remove('')
   return theList

# Same as above, but only get one line of stdout so pipe does not block
def getSingleLineOutputOf(cmd):
   thePipe = os.popen(cmd)
   cmdOut = thePipe.readline()
   thePipe.close()
   if (len(cmdOut) > 0 and cmdOut[-1] == '\n'):
      cmdOut = cmdOut[0:-1]
   return cmdOut

#
# Takes a string, such as "quick.path", and returns just the critical
# part to these scripts, i.e. "quick".
#
def simplifyPath(longPath):
   retval = longPath

   idx = retval.find(".") # if they sent "quick.path"
   if (idx >= 0):
      retval = retval[0:idx]

   idx = retval.find(":") # if they sent "OPUS_DEFS_DIR:quick.path"
   if (idx >= 0):
      retval = retval[idx+1:]

   return retval


#
# Evaluate a file in OPUS_DEFINIIONS_DIR.  Return a full path name.
# Fails if the resulting file does not exist.
#
def evaluateDefinitionsFile(file):
   cmd = "osfile_stretch_file OPUS_DEFINITIONS_DIR:" + file
   if (file.find(":") > 0): # or, if it already includes a stretch
      cmd = "osfile_stretch_file " + file
   retval = getOutputOf(cmd).strip()
   if (os.path.exists(retval)):
      return retval
   else:
      raise IOError, "File not found: " + retval


#
# Utility routines concerning pipeline files
#
# Each version returns a string
#
def getProcFromPplnFileLine(line):
   return getItemFromPplnFileLine(line, 0)

def getPathFromPplnFileLine(line):
   return getItemFromPplnFileLine(line, 1)

def getNodeFromPplnFileLine(line):
   return getItemFromPplnFileLine(line, 2)

def getPKeyFromPplnFileLine(line):
   return getItemFromPplnFileLine(line, 3)

def getItemFromPplnFileLine(line, index):
   items = line.strip().split()
   numItems = len(items)
   rightNum = 0
   if (numItems == 3 or numItems == 4):
      rightNum = 1
   assert rightNum == 1, "Bad line: " + line
   # the 4th field is optional
   if (index == 3 and numItems == 3):
      return "" # this is indexing out of bounds
   # return indexed item
   return items[index]


def readKeyValsFrom(inFileName, startDict, allowOverrides, substitutions):
   """ Reads keyword-value pairs from the given file, adds them to a dict,
   makes any requested text substitutions, and returns the dictionary.
   @param inFileName: input file name
   @type  inFileName: string
   @param startDict: dictionary to add file items to, not modified (or None)
   @type  startDict: dictionary <string:string>
   @param allowOverrides: 1 if overrides are allowed, 0 if not
   @type  allowOverrides: int
   @param substitutions: dict of any text substitutions to make (or None)
   @type  substitutions: dictionary <string:string>
   """
   # init
   retval = {}
   if (startDict != None):
      retval = copy.deepcopy(startDict)

   # read the file
   f = file(inFileName)
   fileLines = f.readlines()  # reads all contents into a list of strings
   f.close()

   # find key-val pair lines and process them
   for line0 in fileLines:
      line = line0.strip()
      if ((len(line) <= 0) or (line.find("#") == 0)):
         continue
      idxWhite = line.index(" ") # raises ValueError
      key = line[:idxWhite]
      val = line[idxWhite:]
      # use the input dict to make any needed substitutions
      if (substitutions != None):
         for subKey in substitutions.keys():
            val = val.replace(subKey, substitutions[subKey])

      # check for a possible override
      if (not allowOverrides):
         assert key not in retval, \
            "ERROR: Trying to override existing item: "+key+" = "+retval[key]

      # now put it in the dict to be returned
      retval[key] = val # add or update
      
   return retval
# end of readKeyValsFrom


#
# readPipelineFile
#
# Returns a list<list<string>>, where the inner list is a list of lines
# in the pipeline file with the same path, and the outer list is a
# collection of all the different path groupings
#
# Throws AssertionError if ANYTHING unexpected happens, or IOError for
# file-reading problems
#

def readPipelineFile(fileName):

   # Read the file into a list of string lines, their format is
   # "process_name path_name node_name"
   pfile = open(fileName, 'r')
   fileLines = pfile.readlines()
   pfile.close()

   # check
   assert fileLines != None, "Empty pipeline file: " + fileName
   assert len(fileLines) > 0, "Empty pipeline file: " + fileName

   commentln = re.compile("#")

   # Now take that list and sort it first by path.  This way, the caller
   # gets the results in chunks of constant path.
   thePathLists = [] # list<list<string>>, the string is pplnfile format
   for line0 in fileLines:

      ## Skip comment lines (# at _beginning_ of line0):
      if commentln.match(line0) : continue     ## don't include comment lines

      line = line0.strip()
      if (len(line) <= 0):
         continue

      path = getPathFromPplnFileLine(line)
      # we have a path from a line.  if its a new path, make a new list for it
      foundAmatch = 0
      if (len(thePathLists) > 0):
         # see if its in an existing list
         for pl in thePathLists:
            # every string in this list uses the same path, so check 1st string
            pathForThisList = getPathFromPplnFileLine(pl[0])
            if (pathForThisList == path):
               foundAmatch = 1
               pl.append(line)
               break
      if (0 == foundAmatch):
         # we haven't seen this path yet
         newPathList = []                 # create new list for this path
         newPathList.append(line)         # add this line to the new list
         thePathLists.append(newPathList) # add the new list to master list

   return thePathLists


#
# Returns 0 if the PSTAT and CONTEXT server are both up and pingable,
# Returns 1 (with output messages) otherwise
#
def checkServersUp():

   # Step 1:
   #
   # Make sure the servers are up.  This script assumes you have
   # already cleanly brough the servers up.  Check both the
   # opus_bb_server and the opus_env_server
   user = pwd.getpwuid(os.getuid())[0]  # os.getlogin() has some issue

   # We will be doing all this with opus_server_monitor -query here, but
   # add an extra verification that the PSTAT bb server and the env_server
   # are running.
   cmd = "/bin/ps -fu " + user + " | /bin/grep 'opus_bb_server.*-t  *PSTAT'"\
         " | /bin/grep -v grep"
   cmdOut = getOutputOf(cmd)
   if cmdOut.find("opus_bb_server") < 0:
      print "checkServersUp: Error: an opus_bb_server is not running"
      return 1

   cmd = "/bin/ps -fu " + user + " | /bin/grep opus_env_server"\
         " | /bin/grep -v grep"
   cmdOut = getOutputOf(cmd)
   if cmdOut.find("opus_env_server") < 0:
      print "checkServersUp: Error: the opus_env_server is not running"
      return 1

   # find the opus_corba_objs file (assumes we're on the same NFS disk!)
   cmd = "osfile_stretch_file OPUS_DEFINITIONS_DIR:opus_corba_objs"
   objsFile = getOutputOf(cmd)
   
   # get the BB id of the opus_env_server
   cmd = "/bin/grep '= *CONTEXT$' " + objsFile
   idContext = ((getOutputOf(cmd).split("="))[0]).strip()

   # get the BB id of the opus_bb_server -t PSTAT
   cmd = "/bin/grep '= *PSTAT$' " + objsFile
   idPstat = ((getOutputOf(cmd).split("="))[0]).strip()

   # now run opus_server_monitor -query to make sure all is well
   cmd = "opus_server_monitor -query"
   query = getOutputOf(cmd)
   envSrvr = mygrep_buf(idContext + "...OBJECT ACTIVE", query)
   if (len(envSrvr) != 1):
      print "checkServersUp: Error: the opus_env_server is not active"
      print "Output of " + cmd + ":\n" + query
      return 1
   pstSrvr = mygrep_buf(idPstat + "...OBJECT ACTIVE", query)
   if (len(pstSrvr) != 1):
      print "checkServersUp: Error: the PSTAT opus_bb_server is not active"
      print "Output of " + cmd + ":\n" + query
      return 1
   inactive = mygrep_buf("OBJECT INACTIVE", query)
   if (len(inactive) > 0):
      print "checkServersUp: Error: there are INACTIVE servers"
      print "Output of " + cmd + ":\n" + query
      return 1
   
   return 0


#
# pstatTestCountInt() - Returns an int, see docstring below
#
def pstatTestCountInt(pstat_test_cmd, catchStderrToo = True):
   """  Call the pstat_test program via the given command allowing retries if
   the stressed servers balk at the first attempt.  This particular function is
   only meant to be called if the --count arg is being used.
   Returns output int.
   """
   MAX_TRIES = 3

   # sanity checks
   assert pstat_test_cmd.find("pstat_test")>=0, \
      "Usage ERROR!  pstatTestCountInt() meant to be used with pstat_test"
   assert pstat_test_cmd.find("--count")>0 or pstat_test_cmd.find("wc -l")>0, \
      "Usage ERROR!  pstatTestCountInt() is only meant to be used with --count"

   # run the cmd
   attempt = 0
   while (attempt < MAX_TRIES):
      output = ""
      if (catchStderrToo):
         output = getOutputOf(pstat_test_cmd).strip()
      else:
         output = getSingleLineOutputOf(pstat_test_cmd).strip()
      assert len(output)>0, "No output found from: "+pstat_test_cmd

      # check the output for spaces (as error messages would have)
      if (output.find(" ")>=0):
         # this was an errored attempt - allow retries
         print "WARNING: pstatTestCountInt got this result: \""+output+"\""
         print "WARNING: pstatTestCountInt will retry"
         time.sleep(2)
         attempt += 1
      else:
         return int(output)

   # this is an error
   raise ValueError, "pstatTestCountInt gave up on: "+pstat_test_cmd



###############################################################################
# History:
# -------- -----  --------  --------------------------------------
# 09/07/05 52948  Sherbert  Add 3 modules for path file reading
#
###############################################################################
##
## Special case usage of evaluateDefinitionsFile, specifically
## to evaluate a path file.
## 
## Determine full unix path name of named path file
##
def getFullPath(pathname):
    ## Determine full path name of pathname, with error checking
    if 2 != len(pathname.split('.')):
        pathname = pathname + '.path'
    pathname = evaluateDefinitionsFile(pathname)
    return(pathname)

###############################################################################
# History:
# -------- -----  --------  --------------------------------------
# 09/07/05 52948  Sherbert  Add 3 modules for path file reading
# 10/07/05 52948  Sherbert  Add info if get_path_item fails
#
###############################################################################
## Determine name/location of path file, then definitions for resource
## checking for errors as we go.  Resource is expected to return a unix path
## value (directory name) as the resource value's existence as a directory is 
## verified herein.
##
## Usage Example: cdbs_input_dir = getDirFmPath("cdbs","cdbs_input_dir")
##
def getDirFmPath(pathname,resource):
    ## Determine full path name of pathname to feed to
    ## get_path_item command, which isn't as friendly as get_resource_values
    pathFile = getFullPath(pathname)
    cmd = "get_path_item "+pathFile+" "+resource
    (status,resourceValue) = commands.getstatusoutput(cmd)
    if (status):
        msg = "Command FAILED: get_path_item "+pathFile+" "+resource
        msg = msg+"\n\t\t"+resourceValue
        raise ValueError, msg
    ## Verify the directory exists
    if os.path.isdir (resourceValue):
        return(resourceValue)
    else:
        return(resourceValue)
        #raise IOError, "File does NOT exist: " + resourceValue <--Don't want to fail

###############################################################################
# History:
# -------- -----  --------  --------------------------------------
# 09/07/05 52948  Sherbert  Add 3 modules for path file reading
# 10/07/05 52948  Sherbert  Add info if get_path_item fails
#
###############################################################################
## Determine name/location of path file, then definitions for resource
## checking for errors as we go.  This resource is expected to return a string
## value (not a directory name, for example).  
##
## Usage Example: cdbs_machine = getStringFmPath("cdbs.path","cdbs_machine")
##
def getStringFmPath(pathname,resource):
    ## Determine full path name of pathname to feed to
    ## get_path_item command, which isn't as friendly as get_resource_values
    pathFile = getFullPath(pathname)
    cmd = "get_path_item "+pathFile+" "+resource
    (status,resourceValue) = commands.getstatusoutput(cmd)
    if (status):
        msg = "Command FAILED: get_path_item "+pathFile+" "+resource
        msg = msg+"\n\t\t"+resourceValue
        raise ValueError, msg
    return(resourceValue)




################################################################################
# History:
# mm/dd/yy PR     User      Description
# -------- -----  --------  ----------------------------------------------------
# 10/11/05 52947  Sherbert  Add from regr_00104/scripts/getInput_00104.py
# 11/01/05 53009  Sherbert  Slight change in verbose messaging
# 04/10/07 53009  Sherbert  Define OK, NOT_OK
#
################################################################################
## 
## Utility routine to copy a list of files from sourceDir to destination
##
def copyFile(sourceDir, fileList, destination):
    import opusutil
    moduleName = 'copyFile subroutine'
    OK = 0
    NOT_OK = 1

    ## Verify that directories are indeed directories
    if not os.path.isdir( sourceDir) :
        msg = sourceDir+"\n\t\tis NOT a directory"
        opusutil.PrintMsg ('F', msg, moduleName)
        sys.exit(NOT_OK)
    if not os.path.isdir( destination) :
        msg = destination+"\n\t\tis NOT a directory"
        opusutil.PrintMsg ('F', msg, moduleName)
        sys.exit(NOT_OK)

    ## First cd to source directory
    msg = "CD to sourceDir: "+sourceDir
    opusutil.PrintMsg('V',msg)
    try:
        os.chdir (sourceDir)
    except OSError, extraInfo:
        msg = "FAILED os.chdir "+sourceDir
        strExtraInfo = str(extraInfo)+msg
        opusutil.PrintMsg ('F', strExtraInfo, moduleName)
        sys.exit(NOT_OK)

    ## Copy a group of files (in fileList) to destination
    msg = "Copy files to "+destination
    opusutil.PrintMsg('V',msg)
    for File in fileList:
        baseFile = os.path.basename(File)
        try:
            msg = "cp "+baseFile
            opusutil.PrintMsg('V',msg)
            shutil.copy( baseFile, destination )
        except IOError, extraInfo:
            strExtraInfo = str(extraInfo)+" shutil.copy FAILED"
            opusutil.PrintMsg ('F', strExtraInfo, moduleName)
            sys.exit(NOT_OK)
    
        ## Try to verify that the file is now in destination
       #msg = "Check existence of: \n\t\t"+destination+os.sep+baseFile
       #opusutil.PrintMsg('V',msg)
        if not os.path.isfile(destination+os.sep+baseFile) :
            msg = destination+os.sep+baseFile+"\n\t\tis NOT a file.  BAD copy?"
            opusutil.PrintMsg('F',msg, moduleName)
            sys.exit(NOT_OK)
        msg = "Open write permissions on existing file\n\t\t"+destination+os.sep+baseFile
        opusutil.PrintMsg('V',msg)
        os.chmod(destination+os.sep+baseFile, 0644)  # rw-r--r-- 

################################################################################
# History:
# mm/dd/yy PR     User      Description
# -------- -----  --------  ----------------------------------------------------
# 10/11/05 52947  Sherbert  Add from regr_00104/scripts/getInput_00104.py
#                           where it was called mkSubInFile
# 10/13/05 52947  Sherbert  Add the missing rename block
# 10/14/05 52947  Sherbert  Remove extraneous blank lines
# 04/10/07 53009  Sherbert  Define OK, NOT_OK
#
################################################################################
##
## Utility routine to perform a string substitution 
## in order to edit a file "in place"
##   inFile is the file in which the substitution will occur
##   inString is the string to substitute
##   inSubVal is the value to substitute for inStrign
##
## throws IOError if file is not found
##
def changeStringInFile(inFile, inString, inSubVal):
    import opusutil
    moduleName = 'changeStringInFile'
    OK = 0
    NOT_OK = 1
   #print "\n"

   #opusutil.PrintMsg ('V', "## inFile is "+inFile)
   #opusutil.PrintMsg ('V', "## inString is "+inString)
   #opusutil.PrintMsg ('V', "## inSubVal is "+inSubVal)
   #opusutil.PrintMsg ('V', "## Try to open "+inFile)

    try:
        inFileFD = file(inFile,'r')
    except IOError, extraInfo:
        strExtraInfo = str(extraInfo)
        opusutil.PrintMsg ('F', strExtraInfo, moduleName)
        sys.exit(NOT_OK)

    ## Need to write lines to a temporary file before rename to real file
    ## os.tmpnam() has security vulnerabilities
    ## os.tmpfile() goes away when closed, I think
    ## os.rename() gives cross-link path error for /tmp/sherbert
    ## Therefore, try using directory in which input file is found
    ##    If inFile exists, then tmpDir will exist
    
    tmpDir  = os.path.dirname(inFile)
    tmpFile = tmpDir+os.sep+"ZZTmpFile.txt"
    opusutil.PrintMsg ('V', "tmpFile is "+tmpFile)
    ## Open temporary file for writing
    try:
        tmpFileFD = file(tmpFile,'w')
    except IOError, extraInfo:
        strExtraInfo = str(extraInfo)
        opusutil.PrintMsg ('F', strExtraInfo, moduleName)
        sys.exit(NOT_OK)

    ## Read each line of input file, replace string when found, write to temp file
    while 1:
        thisLine = inFileFD.readline()
        if not thisLine: break
        thisLine = re.sub(inString, inSubVal, thisLine)
        ## Successful so far, therefore write lines to temp file
        tmpFileFD.write(thisLine)

    inFileFD.close()
    tmpFileFD.close()

    ## Rewrite the input file with substituted lines from the temp file
    ## using the file rename facility
    msg = "rename " + tmpFile + " " + inFile
    opusutil.PrintMsg ("V", msg )
    try:
        os.rename(tmpFile, inFile)
    except OSError, extraInfo:
        strExtraInfo = str(extraInfo)
        opusutil.PrintMsg ('F', strExtraInfo, moduleName)
        sys.exit(NOT_OK)

